from box.box import Box as Box
from typing import Any, Optional

class ConfigBox(Box):
    def __getattr__(self, item: Any): ...
    def __dir__(self): ...
    def bool(self, item: Any, default: Optional[Any] = ...): ...
    def int(self, item: Any, default: Optional[Any] = ...): ...
    def float(self, item: Any, default: Optional[Any] = ...): ...
    def list(self, item: Any, default: Optional[Any] = ..., spliter: str = ..., strip: bool = ..., mod: Optional[Any] = ...): ...  # type: ignore
    def getboolean(self, item: Any, default: Optional[Any] = ...): ...
    def getint(self, item: Any, default: Optional[Any] = ...): ...
    def getfloat(self, item: Any, default: Optional[Any] = ...): ...
    def copy(self): ...
    def __copy__(self): ...
