from box.box import Box as Box

class SBox(Box):
    @property
    def dict(self): ...
    @property
    def json(self): ...
    @property
    def yaml(self): ...
    @property
    def toml(self): ...
    def copy(self): ...
    def __copy__(self): ...

class DDBox(Box): ...
